#include "larccore.h"
#include "larcutils.h"
#include <math.h>
#include <string.h>
#include <time.h>

/*
 * Perform LARC coding of data matrix X in dictionary D.
 */
void larccore(double C[], double Coh[], double D[], double X[], double DtX[], double G[], double mut, mwSize k, double err, int verbose, mwSize d, mwSize l, mwSize n) {

	double *c, *y, *Dty, *r, *L, *D_active, *G_active;
	double *alpha, *alpha_a, *alpha_i, *s, *g, *v, *w, *uw, *u, *a, *c_active;
	double alpha_max, coherence, gamma;
	double wtw, b, t;
	double r2norm;
	mwIndex *active, *inactive;
	mwIndex active_num, inactive_num;
	mwIndex index_max, index_current;
	mwIndex nn, aa;
	clock_t time_start, time_lastprinted;
	int remaining_minutes, remaining_hours;
	double remaining_seconds;

	/* Memory allocations */
	c = (double*)mxMalloc(l*sizeof(double));   /* Sparse coding vector C(:,nn) */
	y = (double*)mxMalloc(d*sizeof(double));   /* Current estimate of X(:,nn) */
	Dty = (double*)mxMalloc(l*sizeof(double)); /* D'*y */
	r = (double*)mxMalloc(d*sizeof(double));   /* Current residual vector */

	active = (mwIndex*)mxMalloc(k*sizeof(mwIndex));   /* Index set of active atoms */
	inactive = (mwIndex*)mxMalloc(l*sizeof(mwIndex)); /* Index set of inactive atoms, i.e. the complement of the active atoms set */

	alpha = (double*)mxMalloc(l*sizeof(double));   /* Residual correlation D'*r */
	alpha_a = (double*)mxMalloc(l*sizeof(double)); /* Residual correlation D(:, active)'*r */
	alpha_i = (double*)mxMalloc(l*sizeof(double)); /* Residual correlation D(:, inactive)'*r */

	s = (double*)mxMalloc(l*sizeof(double)); /* Sign vector, sign(alpha(active)) */
	g = (double*)mxMalloc(l*sizeof(double));
	uw = (double*)mxMalloc(l*sizeof(double)); /* Equiangular weight vector */
	u = (double*)mxMalloc(d*sizeof(double));  /* Equiangular direction */

	a = (double*)mxMalloc(l*sizeof(double)); /* G(:,active)*uw */

	c_active = (double*)mxMalloc(k*sizeof(double)); /* c_active = c(active) */

	L = (double*)mxMalloc(k*k*sizeof(double)); /* Cholesky decomposition L*L' = D(:, active)'*D(:, active) */
	v = (double*)mxMalloc(k*sizeof(double));   /* Used during Cholesky update */
	w = (double*)mxMalloc(k*sizeof(double));   /* Used during Cholesky update */

	D_active = (double*)mxMalloc(d*d*sizeof(double));
	G_active = (double*)mxMalloc(l*l*sizeof(double));

	/* Initialize Timer */
	time_start = clock();
	time_lastprinted = time_start;

	/* LARC code each signal */
	for (nn=0; nn<n; nn++) {

		/* Initializations */
		zeros(c, l); /* c = zeros(l, 1); */
		zeros(y, d);
		zeros(Dty, l);
		mcolr(r, X, nn, d); /* r = X(:, nn) */

		active_num = 0;
		inactive_num = l;
		index_set_complement(inactive, l, active, active_num); /* inactive = {1:l} \ active(1:active_num} */

		while (active_num <= k) {

			/* Find next atom which to add to the active set */
			vadd(alpha, -1, Dty, 1, DtX + nn*l, l);           /* alpha = DtX(:, nn) - Dty), ~ residual coherence */
			vindexr(alpha_i, alpha, inactive, inactive_num);  /* alpha_i = alpha(inactive) */
			index_max = vindexmax(alpha_i, inactive_num);     /* arg max_j |alpha_i(j)|, j in inactive set */
			index_current = inactive[index_max];              /* Index of next atom */
			alpha_max = fabs(alpha[index_current]);

			/* Residual coherence stopping criterion */
			r2norm = sqrt(vdot(r, r, d));
			coherence = alpha_max / r2norm;
			melementw(Coh, active_num, nn, coherence, k);  /* Coh(active_num, nn) = coherence */
			if (coherence < mut) break;                    /* Compare to mu threshold stopping criterion */
			if (r2norm < err) break;                       /* Residual sufficiently small */

			/* Cholesky update of L, based on previous active atom set */
			if (active_num == 0) {
				*L = 1;
			}
			else {
				/* v = D(:, active)' * D(:, index_current), extract from G */
				vindexr(v, G+index_current*l, active, active_num);

				/* Solve L*w = v for w */
				solveltr(w, L, 0, v, active_num, k);

				/* Add w' as new row to L */
				mroww(L, active_num, k, w, active_num);

				/* Complete new row of L with sqrt(1 - w'*w), i.e. set diagonal element */
				wtw = vdot(w, w, active_num);
				if ((1-wtw) < 1e-14) {
					/* Selected atoms are dependent, unable to compute Cholesky decomposition */
					break;
				}
				else {
					melementw(L, active_num, active_num, sqrt(1-wtw), k);
				}
			}

			/* Add newly selected atom to active set */
			active[active_num] = index_current;
			active_num++;
			inactive_num--;
			index_set_complement(inactive, l, active, active_num);

			/* Updates */
			vindexr(alpha_a, alpha, active, active_num);            /* alpha_a = alpha(active) */
			mcolcopy(D_active, active_num-1, D, index_current, d);  /* D_active(:,active_num-1) = D(:,index_current) */
			mcolcopy(G_active, active_num-1, G, index_current, l);  /* G_active(:,active_num-1) = G(:,index_current) */

			/* Compute equiangular direction u */
			vsign(s, alpha_a, active_num);
			solvecholesky(g, L, s, active_num, k);
			b = 1/sqrt(vdot(g, s, active_num));
			vscalec(uw, g, b, active_num);
			mvprod(u, D_active, d, active_num, uw);

			/* Compute step size gamma */
			if (active_num == k) {

				/* Least squares solution */
				gamma = alpha_max/b;
			}
			else {

				mvprod(a, G_active, l, active_num, uw); /* a = G(:, active)*uw */

				/* Find minimum positive value */
				gamma = HUGE_VAL;
				for (aa=0; aa<inactive_num; aa++) {
					t = (alpha_max - alpha[inactive[aa]]) / (b - a[inactive[aa]]);
					if (t > 0) {
						gamma = dmin(gamma, t);
					}
					t = (alpha_max + alpha[inactive[aa]]) / (b + a[inactive[aa]]);
					if (t > 0) {
						gamma = dmin(gamma, t);
					}
				}
				gamma = dmin(gamma, alpha_max/b);
			}

			/* Perform step and update coding vector coefficients */
			vadd(y, 1, y, gamma, u, d);    /* y = y + gamma*u */
			vadd(r, 1, X+nn*d, -1, y, d);  /* r = X(:,nn) - y */

			vindexr(c_active, c, active, active_num);            /* c_active = c(active) */
			vadd(c_active, 1, c_active, gamma, uw, active_num);  /* c_active = c_active + gamma*uw */
			vindexw(c, c_active, active, active_num);            /* c(active) = c_active */

			vadd(Dty, 1, Dty, gamma, a, l);

			/* Stopping criteria */

			/* Number of active atoms */
			if (active_num == k) {
				break;
			}
		
		} /* End While */

		/* Copy coding vector solution to coding matrix */
		mcolw(C, nn, l, c);

		/* Display status messages */
		if (verbose >= 1 && (clock()-time_lastprinted)/(double)CLOCKS_PER_SEC >= verbose) {

			time_lastprinted = clock();

			/* Estimated remaining time */
			sec2hms( ((n-nn-1)/(double)(nn)) * ((time_lastprinted - time_start)/(double)CLOCKS_PER_SEC),
				&remaining_hours, &remaining_minutes, &remaining_seconds);

			mexPrintf("LARC: Signal %d/%d, estimated time remaining (hh:mm:s): %02d:%02d:%05.2f\n",
				nn+1, n, remaining_hours, remaining_minutes, remaining_seconds);
			mexEvalString("drawnow;");
		}

	} /* End For */

	/* Free memory */
	mxFree(c);
	mxFree(y);
	mxFree(Dty);
	mxFree(r);

	mxFree(active);
	mxFree(inactive);

	mxFree(alpha);
	mxFree(alpha_a);
	mxFree(alpha_i);

	mxFree(s);
	mxFree(g);
	mxFree(uw);
	mxFree(u);

	mxFree(a);
	mxFree(c_active);

	mxFree(L);
	mxFree(v);
	mxFree(w);
	mxFree(D_active);
	mxFree(G_active);

	return;
}
