function [C, Coh] = larcmat(D, X, k_max, coh_res)
% LARS Batch coding of data X using dictionary D.
%
% INPUTS
% D: (d x l) dictionary
% X: (d x n) data matrix
%
% OUTPUTS
% C: (l x n) coding of samples
% R: (d x n) residuals

[d,l] = size(D);  % dimensionality of space and number of atoms
n = size(X,2);    % number of observations

if isinf(k_max)
    k_max = min(d-1,l);
end

%% Coding

G = D'*D;  % precompute Gram matrix

C = zeros(l,n);
Coh = zeros(k_max,n);
for nn = 1:n
    [C(:,nn) Coh(:,nn)] = larc_inner(D, G, X(:,nn), k_max, coh_res);
end

end

function [c, coh] = larc_inner(D, G, x, k_max, coh_res)

[d,l] = size(D);
vars = 0;

c = zeros(l,1);       % sparse code
coh = ones(k_max,1);  % residual coherence curve, fill with ones to stop at k_max
y = zeros(d,1);       % approximation to x
r = x;                % residual

Dtx = D'*x;
Dty = zeros(l,1);
active = zeros(1,l) == 1;  % active atoms

while true
    
    % computation of D'*r split in two parts, to avoid computing all inner
    % products every time (see update of Dty)
    alpha = Dtx - Dty;
    
    [alpha_max, max_ndx] = max(abs(alpha(~active)));
    inactive_ndx = find(~active);
    cur = inactive_ndx(max_ndx(1));  % activate first maximum atom
    
    active(cur) = true;
    vars = vars + 1;
    
    % coherence stopping criterion
    coh(vars) = alpha_max / norm(r);  % coherence: maximum absolute correlation
    if coh(vars) < coh_res
        break
    end
    
    % terminate if approximation error too small
    if norm(r) < 1e-6
        break
    end
    
    % Equiangular direction u
    s = sign(alpha(active));  % signs of unnormalized correlations
    S = s*ones(1,vars); 
    [GA1, rcnd] = linsolve(G(active,active).*S'.*S, ones(vars,1));
    if rcnd < 1e-14;
        break;  % break if active set becomes linearly dependent
    end
    AA = 1/sqrt(sum(GA1));
    w = AA*GA1.*s;      % weights applied to each active variable to get equiangular direction
    u = D(:,active)*w;  % equiangular direction
    a = G(:,active)*w;
    
    % Step size gamma
    if vars == min(d-1,l)
        % least squares solution when all atoms active
        gamma = alpha_max/AA;
    else
        temp = [(alpha_max - alpha(~active))./(AA - a(~active)); ...
                (alpha_max + alpha(~active))./(AA + a(~active))];
        gamma = min([temp(temp > 0); alpha_max/AA]);
    end
    
    % Take step
    y = y + gamma*u;  % update estimation of x
    r = x - y;        % update residual
    c(active) = c(active) + gamma*w;  % update sparse code
    Dty = Dty + gamma*a;  % update unnormalized correlation to y
    
    % cardinality stopping criterion
    if vars == k_max
        break
    end
end

end