#include "larcutils.h"

/*
 * y = X(:, n)
 */
void mcolr(double y[], const double X[], mwIndex n, mwSize d) {

	cblas_dcopy(d, X + n*d, 1, y, 1);
}

/*
 * A(:, p) = B(:, q), column dimension d.
 */
void mcolcopy(double A[], mwIndex p, const double B[], mwIndex q, mwSize d) {

	cblas_dcopy(d, B+q*d, 1, A+p*d, 1);
}

/*
 * X(i, j) = v, X has m rows, n columns.
 */
void melementw(double X[], mwIndex i, mwIndex j, double v, mwSize m) {

	*(X + j*m + i) = v;
}

/*
 * X(i, :) = v, X has d rows, l is the length of v.
 */
void mroww(double X[], mwIndex i, mwSize d, const double v[], mwSize l) {

	mwIndex ll;

	for (ll=0; ll<l; ll++) {
		X[ll*d+i] = v[ll];
	}
}

/*
 * X(:, i) = v, X has d rows.
 */
void mcolw(double X[], mwIndex i, mwSize d, const double v[]) {

	mwIndex dd;

	for (dd=0; dd<d; dd++) {
		X[i*d+dd] = v[dd];
	}
}

/*
 * b = A*x, A has m rows, n columns.
 */
void mvprod(double b[], const double A[], mwSize m, mwSize n, const double x[]) {

	cblas_dgemv(CblasColMajor, CblasNoTrans, m, n, 1, A, m, x, 1, 0, b, 1);
}

/*
 * x = zeros(n).
 */
void zeros(double x[], mwSize n) {

	mwIndex ii;

	for (ii=0; ii<n; ii++) {
		x[ii] = 0;
	}
}

/*
 * y = a1*x1 + a2*x2, vectors x1 and x2 have length n, a1 and a2 are scalars.
 */
void vadd(double y[], double a1, const double x1[], double a2, const double x2[], mwSize n) {

	mwIndex nn;

	for (nn=0; nn<n; nn++) {
		y[nn] = a1*x1[nn] + a2*x2[nn];
	}
}

/*
 * Returns the dot product x1'*x2 of vectors x1, x2, both of length n.
 */
double vdot(const double x1[], const double x2[], mwSize n) {

	return cblas_ddot(n, x1, 1, x2, 1);
}

/*
 * Scales vector x of length k with scalar a. x = a*x.
 */
void vscale(double x[], double a, mwSize k) {

	cblas_dscal(k, a, x, 1);
}

/*
 * Scales vector x of length k with scalar a, writes to vector y. y = a*x.
 */
void vscalec(double y[], const double x[], double a, mwSize k) {

	cblas_dcopy(k, x, 1, y, 1);
	cblas_dscal(k, a, y, 1);
}


/*
 * Sets y = x(indices), i.e. accesses sub-vector of x indexed by indices.
 */
void vindexr(double y[], const double x[], const mwIndex indices[], mwSize k) {

	mwSize kk;

	for (kk=0; kk<k; kk++) {

		y[kk] = x[indices[kk]];
	}
}

/*
 * Sets y(indices) = x, i.e. writes the elements of x to the sub-vector of
 * y indexed by indices.
 */
void vindexw(double y[], const double x[], const mwIndex indices[], mwSize k) {

	mwIndex kk;

	for (kk=0; kk<k; kk++) {

		y[indices[kk]] = x[kk];
	}
}

/*
 * Returns the index i for which |x[i]| is max, for all i = 1:k
 */
mwIndex vindexmax(const double x[], mwSize k) {

	mwIndex index_max;

	return index_max = cblas_idamax(k, x, 1);
}

/*
 * Writes to vector s the signs of vector elements of x. Length of x and s is d.
 */
void vsign(double s[], const double x[], mwSize d) {

	mwIndex dd;

	for (dd=0; dd<d; dd++) {

		if (x[dd] >= 0) {
			s[dd] = +1;
		}
		else {
			s[dd] = -1;
		}
	}
}

/*
 * Returns 1 if index a is a set member of index set b(1:k), 0 otherwise.
 */
unsigned short int is_index_set_member(mwIndex a, const mwIndex b[], mwSize k) {

	unsigned short int is_member;
	mwSize kk;

	is_member = 0;

	for (kk=0; kk<k; kk++) {
		if (b[kk] == a) {
			is_member = 1;
			break;
		}
	}

	return is_member;
}

/*
 * Returns the set difference, complement = {1:l} \ set{1:k}
 */
void index_set_complement(mwIndex complement[], mwSize l, const mwIndex set[], mwSize k) {

	mwIndex ll;
	mwIndex ii;

	if (k==0) {

		/* Return the full set {1:l} */
		for (ll=0; ll<l; ll++) {
			complement[ll] = ll;
		}
	}
	else {

		/* Return the set difference */
		ii = 0;
		for (ll=0; ll<l; ll++) {

			if (!is_index_set_member(ll, set, k)) {

				complement[ii] = ll;
				ii++;
			}
		}
	}
}

/*
 * Solve L*w = v for w, where L is lower triangular.
 */
void solveltr(double w[], double L[], int transposed, const double v[], mwSize k, mwSize l) {

	char uplo, trans, diag;
	int n, lda, ldb;
	int nrhs, info;

	double *temp;

	temp = (double*)mxMalloc(k*sizeof(double));

	uplo = 'L';
	if (transposed != 0) {
		trans = 'T';
	}
	else {
		trans = 'N';
	}
	diag = 'N';
	nrhs = 1;

	n = k;
	lda = l;
	ldb = l;

	/* Copy input v to avoid side-effect, since it is overwritten by dtrtrs */
	cblas_dcopy(k, v, 1, temp, 1);

	/* Solve system of equations */
	dtrtrs(&uplo, &trans, &diag, &n, &nrhs, L, &lda, temp, &ldb, &info);

	/* dtrtrs has overwritten v with solution, copy to output w */
	cblas_dcopy(k, temp, 1, w, 1);

	mxFree(temp);
}

/*
 * Solves L*L'*x = b for x by using the cholesky factorization L, lower triangular.
 */
void solvecholesky(double x[], double L[], const double b[], mwSize k, mwSize l) {

	double *temp;

	temp = (double*)mxMalloc(k*sizeof(double));

	solveltr(temp, L, 0, b, k, l);
	solveltr(x, L, 1, temp, k, l);

	mxFree(temp);
}

/*
 * Returns min(a,b)
 */
double dmin(const double a, const double b) {

	if (a < b) {
		return a;
	}
	else {
		return b;
	}
}

/*
 * Converts seconds to hours, minutes and seconds.
 */
void sec2hms(double sec, int *hours, int *minutes, double *seconds) {

	*hours = (int)(floor(sec/3600)+1e-2);
	sec = sec - (*hours)*3600;
	*minutes = (int)(floor(sec/60)+1e-2);
	*seconds = sec - (*minutes)*60;
}
